from django import template

from collections import OrderedDict

from pbs.document.models import DocumentTag, DocumentCategory

register = template.Library()


@register.filter
def tag_id(name):
    """
    Returns a tag id for the given tag name, the lookup is case-insensitive.
    """
    try:
        return DocumentTag.objects.get(name__iexact=name).id
    except DocumentTag.DoesNotExist:
        pass


@register.assignment_tag
def filter_by_tag_name(qs, name):
    """
    Filters the queryset by tag name, the lookup is case-insensitive.
    Ensure that only documents that can be included are included (no zipped
    documents). All image types are automatically converted to pdf, so checking
    for ".pdf" extension should be sufficient.
    """
    return filter(lambda x: x.filename.endswith('.pdf'), qs.tag_names(name))


@register.assignment_tag
def document_names_by_category(prescription):
    """
    Return a dictionary of document names indexed by category.
    Used at the end of Part D to display every document.
    """
    names = OrderedDict()
    for category in DocumentCategory.objects.all().order_by('order'):
        items = []
        for item in prescription.document_set.filter(category=category):
            docs.append(item.document.name.split("/")[-1])
        names[category.name] = items
    return names


@register.inclusion_tag('admin/document_list.html', takes_context=True)
def document_list(context):
    """
    Created for Action report on the CAR screens.
    """
    current = context.get('current')
    request = context.get('request')
    if 'pre_burn' in request.GET:
        document_set = current.document_set.tag_names('Pre Burn Actions List')
        document_type = 'Pre Burn Actions Lists'
    elif 'day_of_burn' in request.GET:
        document_set = current.document_set.tag_names(
            'Day of Burn Actions List')
        document_type = 'Day of Burn Actions Lists'
    elif 'post_burn' in request.GET:
        document_set = current.document_set.tag_names('Post Burn Actions List')
        document_type = 'Post Burn Actions Lists'
    else:
        document_set = current.document_set.all()
        document_type = 'All documents'

    return {
        "current": current,
        "request": request,
        "document_type": document_type,
        "document_set": document_set
    }
